import { Iterable } from '../../../../../lib/src/collection/Iterable.ts'
import ConnectionExecutable from '../ConnectionExecutable.ts'
import {QueryRow} from '../../../db/types.ts'
import {Collection} from '../../../../../lib/src/collection/Collection.ts'

/**
 * Abstract iterable that wraps an executable query.
 * @extends Iterable
 * @abstract
 */
export abstract class ResultSet<T> extends Iterable<any> {

    protected constructor(
        /**
         * The executable query to wrap.
         * @type ConnectionExecutable
         */
        protected executeable: ConnectionExecutable,
    ) {
        super()
    }

    /**
     * Process a single incoming query row to the output format.
     * @param {QueryRow} row
     * @return Promise<any>
     */
    abstract async process_row(row: QueryRow): Promise<T>

    /**
     * Get the result at index i.
     * @param {number} i
     * @return Promise<any>
     */
    async at_index(i: number) {
        return this.process_row(await this.executeable.get_row(i))
    }

    /**
     * Get a collection of results for the given range of rows.
     * @param {number} start
     * @param {number} end
     * @return Promise<Collection>
     */
    async from_range(start: number, end: number) {
        const results = await this.executeable.get_range(start, end)
        const returns = new Collection<T>()
        for ( const row of results ) {
            returns.push(await this.process_row(row))
        }

        return returns
    }

    /**
     * Count the number of results.
     * @return Promise<number>
     */
    async count() {
        return this.executeable.count()
    }
}
