import {Model} from '../Model.ts'
import AppClass from '../../../../lib/src/lifecycle/AppClass.ts'
import ConnectionExecutable from '../../builder/type/ConnectionExecutable.ts'
import {AsyncCollection} from '../../../../lib/src/collection/AsyncCollection.ts'
import {Collection} from '../../../../lib/src/collection/Collection.ts'
import {WhereBuilder} from '../../builder/type/WhereBuilder.ts'
import {RelationBuilder} from './RelationBuilder.ts'
import {Select} from '../../builder/type/Select.ts'
import {FieldSet} from '../../builder/types.ts'
import {Update} from '../../builder/type/Update.ts'
import {Delete} from '../../builder/type/Delete.ts'
import ModelResultOperator from '../ModelResultOperator.ts'
import {ModelSelect} from '../query/ModelSelect.ts'

/**
 * The result of loading a relation.
 */
export type RelationResult<T> = T | Collection<T> | undefined

/**
 * Abstract base class for model-to-model relations.
 * @extends AppClass
 * @abstract
 */
export abstract class Relation<T extends Model<T>, T2 extends Model<T2>> extends AppClass {
    constructor(
        /**
         * The parent model.
         * @type Model
         */
        protected parent: T,
        /**
         * The related model.
         * @type Model
         */
        public readonly related: T2,
    ) {
        super()
    }

    /**
     * Get the value on the parent model that is used in the relation.
     */
    protected abstract get parent_value(): any

    /**
     * Get the result operator for this relation.
     * @return ModelResultOperator
     */
    public get_operator() {
        const related_class = this.related.constructor as typeof Model
        return this.make(ModelResultOperator, related_class)
    }

    /**
     * Get the query instance for this relation.
     * @returns ConnectionExecutable
     * @abstract
     */
    public abstract query(): ConnectionExecutable<T2>

    /**
     * Scope an incoming query to the result set of this relation.
     * @param {WhereBuilder} where
     * @abstract
     */
    public abstract scope_query(where: WhereBuilder): void

    /**
     * Build a query to eager load this relation.
     * @param {ModelSelect} parent_query - the incoming query
     * @param {Collection} result - the loaded parents
     * @return ModelSelect
     * @abstract
     */
    public abstract build_eager_query(parent_query: ModelSelect<T>, result: Collection<T>): ModelSelect<T2>

    /**
     * Match the results from an eager load to only those belonging to this relation's parent.
     * @param {Collection} possibly_related
     * @return Collection
     * @abstract
     */
    public abstract match_results(possibly_related: Collection<T>): Collection<T>

    /**
     * Set the value of this relation, caching it.
     * @param {Collection} related
     * @abstract
     */
    public abstract set_value(related: Collection<T2>): void

    /**
     * Get the cached value of this relation.
     * @return RelationResult
     * @abstract
     */
    public abstract get_value(): Collection<T2> | T2 | undefined

    /**
     * Returns true if this relation has been loaded.
     * @return boolean
     * @abstract
     */
    public abstract is_loaded(): boolean

    /**
     * Fetch the results of this relation, by query.
     * @return AsyncCollection
     */
    public fetch(): AsyncCollection<T2> {
        return this.query().results()
    }

    /**
     * Get the results of this relation from the database.
     * @return Promise<RelationResult>
     */
    public abstract get(): Promise<RelationResult<T2>>

    /**
     * Allow awaiting the relation to get the cached results, or fetched.
     * @param callback
     */
    public then(callback: (result: RelationResult<T2>) => any) {
        if ( this.is_loaded() ) {
            callback(this.get_value() as RelationResult<T2>)
        } else {
            this.get().then(result => {
                if ( result instanceof Collection ) {
                    this.set_value(result as any)
                }

                callback(result)
            })
        }
    }

    /**
     * Get the current value of the relation.
     * @type RelationResult
     */
    public get value(): RelationResult<T2> {
        return this.get_value()
    }

    /**
     * Get the query source of the related model.
     * @type QuerySource
     */
    public get related_query_source() {
        const related_class = this.related.constructor as typeof Model
        return related_class.query_source()
    }

    /**
     * Get an instance of the relation builder for this model.
     * @return RelationBuilder
     */
    public builder(): RelationBuilder<T2> {
        return new RelationBuilder(this)
    }

    /**
     * Get a select statement for this relation.
     * @param {...FieldSet} fields
     * @return Select
     */
    public select(...fields: FieldSet[]): Select<T2> {
        if ( fields.length < 1 ) fields.push(this.related.qualify_column('*'))
        return this.builder().select(...fields)
    }

    /**
     * Get an update statement for this relation.
     * @return Update
     */
    public update(): Update<T2> {
        return this.builder().update()
    }

    /**
     * Get a delete statement for this relation.
     * @return Delete
     */
    public delete(): Delete<T2> {
        return this.builder().delete()
    }
}
